
#include <linux/config.h>

#ifdef CONFIG_MEDUSA

#include <linux/version.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <asm/uaccess.h>

#include <asm/semaphore.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/reboot.h>
#include <linux/malloc.h>

#include "medusa.h"

__u16	medusa_magic=1234;

struct task_struct * constable=NULL;
rwlock_t constable_lock = RW_LOCK_UNLOCKED;

struct wait_queue *constable_wait=NULL;
struct wait_queue *answer_wait=NULL;
static struct semaphore request_mutex=MUTEX;


static int msg_id=1;
static int intercom_status,intercom_istatus;
static struct medusa_packet *intercom,*inbuf,
		cmdbuf,cmdansbuf;
static const char *data_to_constable, *datacmd_to_constable;
static char *data_from_constable, *datacmd_from_constable;

/* temporary variables */
static int no_constable=1;
static int constable_exiting=0;
#ifdef CONFIG_MEDUSA_DEBUG
static int constable_off = 0;
#endif

#define	ICOM_FREE	0
#define	ICOM_VALID	1	/* pripravene na poslanie msg */
#define	ICOM_DATA	2	/* pripravene na poslanie dat */
#define	ICOM_WAIT	3	/* cakanie na prijatie msg */
#define	ICOM_WAITDATA	4	/* cakanie na prijatie dat */
#define	ICOM_CMDANS	5	/* pripravene na poslanie odpovede na cmd */
#define	ICOM_CMDDANS	6	/* pripravene na poslanie dat na cmd */

unsigned long stat_bytes_in,stat_bytes_out,stat_pkt_in,stat_pkt_out;

int medusa_open (struct inode *inode, struct file *file);
int medusa_release (struct inode *inode, struct file *file);
ssize_t medusa_read (struct file *filp, char *buf, size_t count, loff_t *ppos);
ssize_t medusa_write (struct file *filp, const char *buf, size_t count, loff_t *ppos);
unsigned int medusa_poll(struct file * filp, poll_table * wait);

#define MEDUSA_MAJOR	111

struct file_operations fops = {
	NULL,		/* seek */
	medusa_read,	/* read */
	medusa_write,	/* write */
	NULL,		/* readdir */
	medusa_poll,	/* poll */
	NULL,		/* ioctl */
	NULL,		/* mmap */
	medusa_open,	/* open */
	NULL,		/* flush */
	medusa_release,	/* release */
	NULL,		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL,		/* revalidate */
	NULL		/* lock */
};

int medusa_open_medusa(struct inode * inode, struct file * file);
int medusa_open_medstat(struct inode * inode, struct file * file);

int medusa_open( struct inode *inode, struct file *file )
{
	
	printk(KERN_DEBUG "medusa: open(): minor=%x\n",MINOR(inode->i_rdev));
	switch (MINOR(inode->i_rdev)) {
		case 0:
#ifdef CONFIG_MEDUSA_DEBUG
		case 2:
			constable_off = (MINOR(inode->i_rdev) != 0);
#endif
			return medusa_open_medusa(inode, file);
		case 1:
			return medusa_open_medstat(inode, file);
	}
	return -ENODEV;
}

ssize_t medusa_read_medstat(struct file *filp, char *buf, size_t count, loff_t *ppos);
int medusa_release_medstat( struct inode *inode, struct file *file );

struct file_operations fops_medstat = {
	NULL,		/* seek */
	medusa_read_medstat,	/* read */
	NULL,	/* write */
	NULL,		/* readdir */
	NULL,	/* poll */
	NULL,		/* ioctl */
	NULL,		/* mmap */
	medusa_open_medstat,	/* open */
	NULL,		/* flush */
	medusa_release_medstat,	/* release */
	NULL,		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL,		/* revalidate */
	NULL		/* lock */
};

int medusa_open_medstat(struct inode * inode, struct file * file)
{
	file->f_op=&fops_medstat;
	return 0;
}

int medusa_release_medstat( struct inode *inode, struct file *file )
{	return(0);
}

ssize_t medusa_read_medstat(struct file *filp, char *buf, size_t count, loff_t *ppos)
{
	static char tmp[160];
	int len;
	
	len = sprintf(tmp, "bytes-in packets-in bytes-out packets-out msg-id\n"
			"%lu %lu %lu %lu %u\n",
			stat_bytes_in, stat_pkt_in,
			stat_bytes_out, stat_pkt_out,
			msg_id
	);
	count = ((len-*ppos)<count) ? (len-*ppos) : count;
	if (*ppos>=len || count<=0 )
		return 0;
	if (copy_to_user(buf,tmp+*ppos,count))
		return -EFAULT;
	*ppos+=count;
	return (count);
}

int medusa_open_medusa(struct inode * inode, struct file * file)
{
	int i;

	if( current->uid!=0 )	/* doplnit luid ... !!! */
		return -EPERM;

	write_lock(&constable_lock);
	if (constable) {
		write_unlock(&constable_lock);
		return -EPERM;
	}
	request_mutex=MUTEX;
	constable=current;
	stat_bytes_in=stat_bytes_out=stat_pkt_in=stat_pkt_out=0l;

	intercom_status=ICOM_FREE;
	intercom_istatus=ICOM_WAIT;
	medusa_magic++;
	printk(KERN_INFO "medusa: Constable registred (pid=%d).\n",current->pid);
	no_constable=0;
	current->med.med_act= 0;
	current->med.med_iact= 0;
	for(i=0;i<NR_syscalls/32;i++)   current->med.med_syscall[i]=0;
	write_unlock(&constable_lock);
	return 0;
}

int medusa_release( struct inode *inode, struct file *file )
{
	if (constable != current)
		return 0;

	constable_exiting=1; /* stop the growth of number of processes
				waiting for 'answer_wait' queue */

	up (&request_mutex); /* start all waiting processes */

	if (inbuf)
		inbuf->answer=MED_ERR;
	if (answer_wait)
		wake_up(&answer_wait);
	write_lock(&constable_lock);

	constable=NULL;
	constable_exiting=0;
	printk(KERN_INFO "medusa: Constable unregistred.\n");
	write_unlock(&constable_lock);

#ifdef CONFIG_MEDUSA_HALT
	notifier_call_chain(&reboot_notifier_list, SYS_HALT, NULL);
	printk(KERN_EMERG "medusa: Security problem, system halted.\n");
	machine_halt();
	do_exit(0);
#endif CONFIG_MEDUSA_HALT
#ifdef CONFIG_MEDUSA_REBOOT
	notifier_call_chain(&reboot_notifier_list, SYS_RESTART, NULL);
	printk(KERN_EMERG "medusa: Security problem, restarting system.\n");
	machine_restart(NULL);
#endif CONFIG_MEDUSA_REBOOT
	return 0;
}

ssize_t medusa_read(struct file *filp, char *buf, size_t count, loff_t *ppos)
{
	if( constable!=current )
		return -EPERM;
	if( !access_ok(VERIFY_WRITE, buf, count) )
		return -EFAULT;
	
	if( intercom_istatus==ICOM_CMDANS )
	{	if( count != sizeof(struct medusa_packet) )
			return -EIO;
		cmdansbuf.id=0;
		__copy_to_user(buf,&cmdansbuf,count);
		stat_bytes_out+=count; stat_pkt_out++;
		if(datacmd_to_constable!=NULL)	intercom_istatus=ICOM_CMDDANS;
		else				intercom_istatus=ICOM_WAIT;
		return( count );
	}
	else if( intercom_istatus==ICOM_CMDDANS )
	{	if( count != cmdansbuf.data_len )
			return -EIO;
		__copy_to_user(buf,datacmd_to_constable,count);
		stat_bytes_out+=count;
		intercom_istatus=ICOM_WAIT;
		return( count );
	}
	while( intercom_status==ICOM_FREE )
	{	if( count != sizeof(struct medusa_packet) )
			return -EIO;
		sleep_on(&constable_wait);
	}
	if( intercom_status==ICOM_VALID )
	{	if( count != sizeof(struct medusa_packet) )
			return -EIO;
		__copy_to_user(buf,intercom,count);
		stat_bytes_out+=count; stat_pkt_out++;
		if( data_to_constable!=NULL )	intercom_status=ICOM_DATA;
		else				intercom_status=ICOM_WAIT;
		return( count );
	}
	else if( intercom_status==ICOM_DATA )
	{	if( count != intercom->data_len )
			return -EIO;
		__copy_to_user(buf,data_to_constable,count);
		stat_bytes_out+=count;
		intercom_status=ICOM_WAIT;
		return( count );
	}
	return -EIO;
}

ssize_t medusa_write(struct file *filp, const char *buf, size_t count, loff_t *ppos)
{
	if( constable!=current )
		return -EPERM;
	if( !access_ok(VERIFY_READ, buf, count) )
		return -EFAULT;
	if( intercom_istatus==ICOM_WAIT )
	{	if( count != sizeof(struct medusa_packet) )
			return -EIO;
		__copy_from_user(&cmdbuf,buf,count);
		stat_bytes_in+=count; stat_pkt_in++;
		datacmd_from_constable=NULL;
		if( cmdbuf.id != 0 )	/* answer */
		{	if( intercom_status!=ICOM_WAIT
				|| cmdbuf.id!=intercom->id )
				return -EIO;
			memcpy(inbuf,&cmdbuf,count);
		}
		if( cmdbuf.answer==MED_ERR )	cmdbuf.data_len=0;
		if( cmdbuf.data_len>0 )	intercom_istatus=ICOM_WAITDATA;
		else			goto Complet;
		return( count );
	}
	else if( intercom_istatus==ICOM_WAITDATA )
	{	if( count != cmdbuf.data_len )
			return -EIO;
		datacmd_from_constable=kmalloc(count, GFP_KERNEL);
		if( datacmd_from_constable==NULL )
			return -ENOMEM;
                __copy_from_user(datacmd_from_constable,buf,count);	
		stat_bytes_in+=count;
Complet:	if( cmdbuf.id==0 ) /* command */
		{
			datacmd_to_constable=medusa_cmd(&cmdbuf,
				datacmd_from_constable,count,&cmdansbuf);
			if (datacmd_from_constable) /* deleted in medusa_cmd */
				datacmd_from_constable = NULL;
			intercom_istatus=ICOM_CMDANS;
			return( count );
		}
		if( intercom_status!=ICOM_WAIT )	/* len pre istotu */
			return -EIO;
		intercom_istatus=ICOM_WAIT;
		data_from_constable=datacmd_from_constable;
		intercom_status=ICOM_FREE;
		wake_up(&answer_wait);
		return( count );
	}
	return -EIO;
}

unsigned int medusa_poll(struct file * filp, poll_table * wait)
{
	if( constable!=current )
		return -EPERM;
	poll_wait(filp, &constable_wait, wait);
	if( intercom_status!=ICOM_FREE )
		return( POLLIN | POLLRDNORM );
	return(0);
}

char *medusa_request( struct medusa_packet *msg, const char *data )
{
	char * retval;
	
	if (msg==NULL) {
		printk(KERN_ERR "medusa: NULL msg in request!\n");
		return NULL;
	}

	if (constable_exiting) { /* speed-up only */
		msg->answer=MED_ERR;
		return NULL;
	}
#ifdef CONFIG_MEDUSA_DEBUG
	if (constable_off)
		return NULL;
#endif

	read_lock(&constable_lock);

	if (constable==NULL) {
		if (!no_constable) {
			printk(KERN_WARNING "medusa: No constable process!\n");
			no_constable=1;
		}
		msg->answer=MED_ERR;
		read_unlock(&constable_lock);
		return NULL;
	}
	if (constable==current) {
		msg->answer=MED_ERR;
		read_unlock(&constable_lock);
		return NULL;
	}
	down(&request_mutex);
	if (constable_exiting) {
		msg->answer=MED_ERR;
		up(&request_mutex);
		read_unlock(&constable_lock);
		return NULL;
	}
	data_from_constable=NULL;
	data_to_constable=data;
	msg->id=msg_id++;
	if(msg_id==0)
		msg_id++;
	intercom=inbuf=msg;
	intercom_status=ICOM_VALID;
	
	read_unlock(&constable_lock);

	wake_up(&constable_wait);
	sleep_on(&answer_wait);

	intercom_status=ICOM_FREE;
	retval=data_from_constable;

	up(&request_mutex);
	return retval;
}

void medusa_putdata( char *data )
{	if( data!=NULL && !IS_ERR(data) )
		kfree(data);
}

int medusa_init(void)
{
	if (register_chrdev(MEDUSA_MAJOR, "medusa", &fops)) {
		printk(KERN_ERR "medusa: unable to get major %d for medusa devices\n",
			MEDUSA_MAJOR);
		return(-EIO);
	}
	printk(KERN_INFO "Medusa DS9 Security System "
			"(http://fornax.elf.stuba.sk/medusa)\n");
	printk(KERN_INFO "       by  Marek Zelem <marek@fornax.elf.stuba.sk>\n");
	printk(KERN_INFO "       and Milan Pikula <www@fornax.elf.stuba.sk>\n");
	medusa_process_init(~0);
	return 0;
}

#endif CONFIG_MEDUSA

