
#include <linux/config.h>

#ifdef CONFIG_MEDUSA

#include <linux/sched.h>
#include <linux/wrapper.h>
#include <linux/interrupt.h>
#include <linux/malloc.h>

#define	EXPORT_SYMTAB
#include <linux/module.h>

#include "medusa.h"

extern struct task_struct * constable;

void medusa_set_proc_inf( struct m_proc_inf *pi, struct task_struct *task)
{
	pi->pid=task->pid;
	pi->parent_pid=pi->child_pid=pi->sibling_pid=0;
	if( task->p_pptr )	pi->parent_pid=task->p_pptr->pid;
	if( task->p_cptr )	pi->child_pid=task->p_cptr->pid;
	if( task->p_osptr )	pi->sibling_pid=task->p_osptr->pid;
	pi->pgrp=task->pgrp;
	pi->uid=task->uid;
	pi->euid=task->euid;
	pi->suid=task->suid;
	pi->fsuid=task->fsuid;
	pi->gid=task->gid;
	pi->egid=task->egid;
	pi->sgid=task->sgid;
	pi->fsgid=task->fsgid;
	pi->cap_effective=task->cap_effective;
	pi->cap_inheritable=task->cap_inheritable;
	pi->cap_permitted=task->cap_permitted;
	pi->luid=task->luid;
	pi->mp=task->med;
}

void medusa_set_proc( struct task_struct *task, struct m_proc_inf *pi )
{
struct med_proc tmp_med;
	if( pi->pid==task->pid )
	{
//		task->pgrp=pi->pgrp;
		if( task->uid != pi->uid )
		{	free_uid(task);
			task->uid=pi->uid;
			alloc_uid(task);
		}
		task->euid=pi->euid;
		task->suid=pi->suid;
		task->fsuid=pi->fsuid;
		task->gid=pi->gid;
		task->egid=pi->egid;
		task->sgid=pi->sgid;
		task->fsgid=pi->fsgid;
		task->cap_effective=pi->cap_effective;
		task->cap_inheritable=pi->cap_inheritable;
		task->cap_permitted=pi->cap_permitted;
//		task->luid=pi->luid;
		tmp_med=pi->mp;
#ifdef CONFIG_MEDUSA_FORCE
		tmp_med.force_code=task->med.force_code;
		tmp_med.force_len=task->med.force_len;
#endif
		task->med=tmp_med;
	}
}


int medusa_pact( int act, struct task_struct *task, __u32 info1, __u32 info2 )
{ struct medusa_packet msg;
  char *s;
	msg.cmd=MED_PACT;
	msg.u.r_pact.act=act;
	medusa_set_proc_inf(&(msg.u.r_pact.proc),current);
	medusa_set_proc_inf(&(msg.u.r_pact.target),task);
	msg.u.r_pact.info1=info1;
	msg.u.r_pact.info2=info2;
	msg.data_len=0;
	s=medusa_request(&msg,NULL);
	if( s!=NULL )
	{	medusa_putdata(s);
		printk("medusa: Invalid incomming data!\n");
	}
	if( msg.answer != MED_ERR )
		medusa_set_proc(current,&(msg.u.a_pact.proc));
	if( msg.answer == MED_NOT )
		return -EPERM;
	if( msg.answer == MED_YES )
		return(0);
	return(1);
}

int medusa_sendsig(int sig, struct task_struct *p)
{ int r;
	if (in_interrupt() || current->pid <= 1 || current == constable)
		return 1;
	if( sig==SIGCHLD )
		return 1;
	if( sig>0 && current->med.med_act & MPACT_KILL )
	{	r=medusa_pact(MPACT_KILL,p,(__u32)sig,0);
		if( r<=0 )	return(r);
	}
	if( p->pid<=1 )		return(1);
	if( !(current->med.med_vs & p->med.med_vs) )
		return -ESRCH;
	return 1;	/* nic */
}

int medusa_see(struct task_struct *p)
{
	if( current->pid<=1 )		return(1);
	if( current==constable )	return(1);
	if( p->pid<=1 )			return(1);
	if( !(current->med.med_vs & p->med.med_vs) )
		return 0;	/* nevidi */
	return 1;		/* vidi */
}

void medusa_setluid( uid_t uid )
{
	if( current->med.med_act & MPACT_SETUID )
		medusa_pact(MPACT_SETUID,current,(__u32)uid,0);
	if( uid==(uid_t)-1 && current->uid!=uid )	return; /* nemenil */
	if( current->luid == (uid_t)-1 )
	{	if( current->uid == (uid_t)-1 )	/* nemalo by nastat */
		{	current->uid= (uid_t)-2;
			printk("medusa: setuid(-1) -> setuid(-2)\n");
		}
		current->luid=current->uid;
	}
}

int medusa_capable( int cap )
{
	if( current->med.med_act & MPACT_CAP )
	{       if( medusa_pact(MPACT_CAP,current,(__u32)CAP_TO_MASK(cap),(__u32)0)==0 )
			return 1;
	}
	return(0);
}

int medusa_fork( struct task_struct *new, unsigned long clone_flags )
{
	if (current->med.force_code)
		new->med.force_code = kmalloc(new->med.force_len, GFP_KERNEL);
	if (!new->med.force_code)
		new->med.force_len = 0;
	else
		memcpy(new->med.force_code,current->med.force_code,
				new->med.force_len);

	if( current->med.med_act & MPACT_FORK )
	{       if( medusa_pact(MPACT_FORK,new,(__u32)clone_flags,(__u32)0)<=0 )
			return -EPERM;
	}
	return(1);
}

void medusa_process_init( __u16 act )
{ int i;
	current->med.med_vs=BEGIN_VS;
	current->med.med_vs_sw=BEGIN_VS;
	current->med.med_m_vs=BEGIN_VS;
	current->med.med_w_vs=BEGIN_VS;
	/* MPACT_MAY_CD nie, lebo je to len pomocna vec */
	current->med.med_act= act & (__u16)~((__u16)MPACT_MAY_CD);
	current->med.med_iact= 0;
	current->med.med_user= 0;
#ifdef CONFIG_MEDUSA_SYSCALL
	for(i=0;i<NR_syscalls/32;i++)	current->med.med_syscall[i]=~0;
#endif
#ifdef CONFIG_MEDUSA_FORCE
	current->med.force_code = NULL;
	current->med.force_len = 0;
#endif
	current->luid=(uid_t)-1;
}

void medusa_kernel_thread( int (*fn)(void *) )
{
//	if( fn==exec_modprobe )
//	{	medusa_process_init(0);
//		printk("EXECUTING modprobe\n");
//	}
//	else
//	{	
//		printk("medusa: thread: fn=%p\n",fn);
		medusa_process_init(MPACT_EXEC|MPACT_FORK);
//	}
}

EXPORT_SYMBOL(medusa_capable);

#endif CONFIG_MEDUSA

